#
# This module contains most of the code of stdeb.
#
import calendar
import os
import re
import select
import shutil
import sys
import time
import codecs
import configparser as ConfigParser
import subprocess
import tempfile
import stdeb  # noqa: F401
from stdeb import log, __version__ as __stdeb_version__

if hasattr(os, 'link'):
    link_func = os.link
else:
    # matplotlib deletes link from os namespace, expected distutils workaround
    link_func = shutil.copyfile

__all__ = ['DebianInfo', 'build_dsc', 'expand_tarball', 'expand_zip',
           'stdeb_cmdline_opts', 'stdeb_cmd_bool_opts', 'recursive_hardlink',
           'apply_patch', 'repack_tarball_with_debianized_dirname',
           'expand_sdist_file', 'stdeb_cfg_options']

DH_MIN_VERS = '12'  # Fundamental to stdeb >= 0.11
DH_DEFAULT_VERS = 12

# Choose the oldest from Debian oldoldstable and currently supported Ubuntu LTS
PYTHON_ALL_MIN_VERS = '2.7.16-1'
PYTHON3_ALL_MIN_VERS = '3.7.3-1'

try:
    # Python 2.x
    from exceptions import Exception
except ImportError:
    # Python 3.x
    pass


class CalledProcessError(Exception):
    pass


class CantSatisfyRequirement(Exception):
    pass


def check_call(*popenargs, **kwargs):
    retcode = subprocess.call(*popenargs, **kwargs)
    if retcode == 0:
        return
    raise CalledProcessError(retcode)


if sys.version_info[0] == 2:
    help_str_py2 = 'If True, build package for python 2. (Default=True).'
    help_str_py3 = 'If True, build package for python 3. (Default=False).'
else:
    assert sys.version_info[0] == 3
    help_str_py2 = 'If True, build package for python 2. (Default=False).'
    help_str_py3 = 'If True, build package for python 3. (Default=True).'

stdeb_cmdline_opts = [
    ('dist-dir=', 'd',
     "directory to put final built distributions in (default='deb_dist')"),
    ('patch-already-applied', 'a',
     'patch was already applied (used when py2dsc calls sdist_dsc)'),
    ('default-distribution=', None,
     "deprecated (see --suite)"),
    ('suite=', 'z',
     "distribution name to use if not specified in .cfg (default='unstable')"),
    ('default-maintainer=', None,
     'deprecated (see --maintainer)'),
    ('maintainer=', 'm',
     'maintainer name and email to use if not specified in .cfg '
     '(default from setup.py)'),
    ('debian-version=', None, 'debian version (Default: 1)'),
    ('extra-cfg-file=', 'x',
     'additional .cfg file (in addition to stdeb.cfg if present)'),
    ('patch-file=', 'p',
     'patch file applied before setup.py called '
     '(incompatible with file specified in .cfg)'),
    ('patch-level=', 'l',
     'patch file applied before setup.py called '
     '(incompatible with file specified in .cfg)'),
    ('patch-posix', 'q',
     'apply the patch with --posix mode'),
    ('remove-expanded-source-dir', 'r',
     'remove the expanded source directory'),
    ('ignore-install-requires', 'i',
     'ignore the requirements from requires.txt in the egg-info directory'),
    ('pycentral-backwards-compatibility=', None,
     'This option has no effect, is here for backwards compatibility, and may '
     'be removed someday.'),
    ('workaround-548392=', None,
     'This option has no effect, is here for backwards compatibility, and may '
     'be removed someday.'),
    ('no-backwards-compatibility', None,
     'This option has no effect, is here for backwards compatibility, and may '
     'be removed someday.'),
    ('guess-conflicts-provides-replaces=', None,
     'If True, attempt to guess Conflicts/Provides/Replaces in debian/control '
     'based on apt-cache output. (Default=False).'),
    ('with-python2=', None,
     help_str_py2),
    ('with-python3=', None,
     help_str_py3),
    ('no-python2-scripts=', None,
     'If True, do not install scripts for python 2. (Default=False).'),
    ('no-python3-scripts=', None,
     'If True, do not install scripts for python 3. (Default=False).'),
    ('force-x-python3-version', None,
     'Override default minimum python3:any dependency with value from '
     'x-python3-version'),
    ('allow-virtualenv-install-location', None,
     'Allow installing into /some/random/virtualenv-path'),
    ('compat=', 'c',
     'Debian compatibility level (Default={})'.format(DH_DEFAULT_VERS)),
    ('with-dh-virtualenv', None,
     'Use dh_virtualenv to build the package instead of python_distutils. '
     'This is useful to embed an application with all its dependencies '
     'and dont touch to the system libraries.'),
    ('with-dh-systemd', None,
     'Adds the systemd addon that will dh_systemd_enable and'
     'dh_systemd_start helpers at the correct time during build.'),
    ('sign-results', None,
     'Use gpg to sign the resulting .dsc and .changes file'),
    ('sign-key=', None,
     'Specify signing key'),
    ('ignore-source-changes', None,
     'Ignore all changes on source when building source package (add -i.* '
     'option to dpkg-source)'),
    ]

# old entries from stdeb.cfg:

# These should be settable as distutils command options, but in case
# we want to support other packaging methods, they should also be
# settable outside distutils. Consequently, we keep the ability to
# parse ConfigParser files (specified with --extra-cfg-file). TODO:
# Also, some (most, in fact) of the above options should also be
# settable in the ConfigParser file.

stdeb_cfg_options = [
    # With defaults
    ('source=', None,
     'debian/control Source: (Default: <source-debianized-setup-name>)'),
    ('package=', None,
     'debian/control Package: (Default: python-<debianized-setup-name>)'),
    ('package3=', None,
     'debian/control Package: (Default: python3-<debianized-setup-name>)'),
    ('suite=', None,
     'suite (e.g. stable, lucid) in changelog (Default: unstable)'),
    ('suite3=', None,
     'suite3 Suites used when building with python3 only (Default: '
     '<same-as-suite>)'),
    ('maintainer=', None,
     'debian/control Maintainer: (Default: <setup-maintainer-or-author>)'),
    ('debian-version=', None, 'debian version (Default: 1)'),
    ('section=', None, 'debian/control Section: (Default: python)'),

    # With no defaults
    ('epoch=', None, 'version epoch'),
    ('forced-upstream-version=', None, 'forced upstream version'),
    ('upstream-version-prefix=', None, 'upstream version prefix'),
    ('upstream-version-suffix=', None, 'upstream version suffix'),
    ('uploaders=', None, 'uploaders'),
    ('copyright-file=', None, 'copyright file'),
    ('build-depends=', None, 'debian/control Build-Depends:'),
    ('build-conflicts=', None, 'debian/control Build-Conflicts:'),
    ('stdeb-patch-file=', None, 'file containing patches for stdeb to apply'),
    ('stdeb-patch-level=', None, 'patch level provided to patch command'),
    ('depends=', None, 'debian/control Depends:'),
    ('depends3=', None, 'debian/control Depends:'),
    ('suggests=', None, 'debian/control Suggests:'),
    ('suggests3=', None, 'debian/control Suggests:'),
    ('recommends=', None, 'debian/control Recommends:'),
    ('recommends3=', None, 'debian/control Recommends:'),
    ('xs-python-version=', None, 'debian/control XS-Python-Version:'),
    ('x-python3-version=', None, 'debian/control X-Python3-Version:'),
    ('dpkg-shlibdeps-params=', None, 'parameters passed to dpkg-shlibdeps'),
    ('conflicts=', None, 'debian/control Conflicts:'),
    ('conflicts3=', None, 'debian/control Conflicts:'),
    ('breaks=', None, 'debian/control Breaks:'),
    ('breaks3=', None, 'debian/control Breaks:'),
    ('provides=', None, 'debian/control Provides:'),
    ('provides3=', None, 'debian/control Provides3:'),
    ('replaces=', None, 'debian/control Replaces:'),
    ('replaces3=', None, 'debian/control Replaces3:'),
    ('mime-desktop-files=', None, 'MIME desktop files'),
    ('mime-file=', None, 'MIME file'),
    ('shared-mime-file=', None, 'shared MIME file'),
    ('setup-env-vars=', None, 'environment variables passed to setup.py'),
    ('udev-rules=', None, 'file with rules to install to udev'),
    ('python2-depends-name=', None,
     'Python 2 Debian package name used in ${python:Depends}'),
    ('dh-python3-params=', None, 'parameters passed to dh_python3'),
    ]

stdeb_cmd_bool_opts = [
    'patch-already-applied',
    'remove-expanded-source-dir',
    'patch-posix',
    'ignore-install-requires',
    'no-backwards-compatibility',
    'force-x-python3-version',
    'allow-virtualenv-install-location',
    'with-dh-virtualenv',
    'with-dh-systemd',
    'sign-results',
    'ignore-source-changes',
    ]


class NotGiven:
    pass


def process_command(args, cwd=None):
    if not isinstance(args, (list, tuple)):
        raise RuntimeError("args passed must be in a list")
    check_call(args, cwd=cwd)


def recursive_hardlink(src, dst):
    dst = os.path.abspath(dst)
    orig_dir = os.path.abspath(os.curdir)
    os.chdir(src)
    try:
        for root, dirs, files in os.walk(os.curdir):
            for file in files:
                fullpath = os.path.normpath(os.path.join(root, file))
                dirname, fname = os.path.split(fullpath)
                dstdir = os.path.normpath(os.path.join(dst, dirname))
                if not os.path.exists(dstdir):
                    os.makedirs(dstdir)
                newpath = os.path.join(dstdir, fname)
                if os.path.exists(newpath):
                    if os.path.samefile(fullpath, newpath):
                        continue
                    else:
                        os.unlink(newpath)
                # print 'linking %s -> %s'%(fullpath,newpath)
                link_func(fullpath, newpath)
    finally:
        os.chdir(orig_dir)


def debianize_name(name):
    "make name acceptable as a Debian (binary) package name"
    name = name.replace('_', '-')
    name = name.lower()
    return name


def source_debianize_name(name):
    "make name acceptable as a Debian source package name"
    name = name.replace('_', '-')
    name = name.replace('.', '-')
    name = name.lower()
    return name


def debianize_version(name):
    "make name acceptable as a Debian package name"
    # XXX should use setuptools' version sorting and do this properly:
    name = name.replace('.dev', '~dev')

    name = name.lower()
    return name


def dpkg_compare_versions(v1, op, v2):
    args = ['/usr/bin/dpkg', '--compare-versions', v1, op, v2]
    cmd = subprocess.Popen(args)
    returncode = cmd.wait()
    if returncode:
        return False
    else:
        return True


def get_cmd_stdout(args):
    cmd = subprocess.Popen(args, stdout=subprocess.PIPE)
    returncode = cmd.wait()
    if returncode:
        log.error('ERROR running: %s', ' '.join(args))
        raise RuntimeError('returncode %d', returncode)
    return cmd.stdout.read()


def get_date_822():
    """return output of 822-date command"""
    source_date_epoch = os.environ.get('SOURCE_DATE_EPOCH', None)
    if source_date_epoch:
        t = time.gmtime(int(source_date_epoch))
    else:
        t = time.localtime()
    return time.strftime(
        calendar.day_abbr[t.tm_wday] + ", %d " +
        calendar.month_abbr[t.tm_mon] + " %Y %H:%M:%S %z", t)


def get_version_str(pkg):
    args = ['/usr/bin/dpkg-query', '--show', '--showformat=${Version}', pkg]
    stdout = get_cmd_stdout(args)
    return stdout.strip().decode('ascii')


def load_module(name, fname):
    import imp

    suffix = '.py'
    found = False
    for description in imp.get_suffixes():
        if description[0] == suffix:
            found = True
            break
    assert found

    fd = open(fname, mode='r')
    try:
        module = imp.load_module(name, fd, fname, description)
    finally:
        fd.close()
    return module


def get_deb_depends_from_setuptools_requires(requirements, on_failure="warn"):
    """
    Suppose you can't confidently figure out a .deb which satisfies a given
    requirement.  If on_failure == 'warn', then log a warning.  If on_failure
    == 'raise' then raise CantSatisfyRequirement exception.  If on_failure ==
    'guess' then guess that python-$FOO will satisfy the dependency and that
    the Python version numbers will apply to the Debian packages (in addition
    to logging a warning message).
    """
    assert on_failure in ("raise", "warn", "guess"), on_failure

    import pkg_resources

    depends = []  # This will be the return value from this function.

    parsed_reqs = []

    for extra, reqs in pkg_resources.split_sections(requirements):
        if extra:
            continue
        parsed_reqs.extend(pkg_resources.parse_requirements(reqs))

    if not parsed_reqs:
        return depends

    if not os.path.exists('/usr/bin/apt-file'):
        raise ValueError('apt-file not in /usr/bin. Please install '
                         'with: sudo apt-get install apt-file')

    # Ask apt-file for any packages which have a .egg-info file by
    # these names.

    # Note that apt-file appears to think that some packages
    # e.g. setuptools itself have "foo.egg-info/BLAH" files but not a
    # "foo.egg-info" directory.

    egginfore = (r"(/(%s)(?:-[^/]+)?(?:-py[0-9]\.[0-9.]+)?\.egg-info)"
                 % '|'.join(
                     req.project_name.replace('-', '_')
                     for req in parsed_reqs))

    args = ["apt-file", "search", "--ignore-case", "--regexp", egginfore]

    if 1:
        # do dry run on apt-file
        dry_run_args = args[:] + ['--dummy', '--non-interactive']
        cmd = subprocess.Popen(dry_run_args, stderr=subprocess.PIPE)
        returncode = cmd.wait()
        if returncode:
            err_output = cmd.stderr.read()
            raise RuntimeError('Error running "apt-file search": ' +
                               err_output.strip())

    try:
        cmd = subprocess.Popen(args, stdin=subprocess.PIPE,
                               stdout=subprocess.PIPE,
                               universal_newlines=True)
    except Exception as le:
        # TODO: catch rc=1 and "E: The cache directory is empty. You need to
        # run 'apt-file update' first.", and tell the user to follow those
        # instructions.
        log.error('ERROR running: %s', ' '.join(args))
        raise RuntimeError('exception %s from subprocess %s' % (le, args))
    returncode = cmd.wait()
    if returncode:
        log.error('ERROR running: %s', ' '.join(args))
        raise RuntimeError('returncode %d from subprocess %s' % (returncode,
                                                                 args))

    inlines = cmd.stdout.readlines()

    dd = {}  # {pydistname: {pydist: set(debpackagename)}}
    E = re.compile(egginfore, re.I)
    D = re.compile("^([^:]*):", re.I)
    eggsndebs = set()
    for line in inlines:
        if line:
            emo = E.search(line)
            assert emo, line
            dmo = D.search(line)
            assert dmo, line
            eggsndebs.add((emo.group(1), dmo.group(1)))

    for (egginfo, debname) in eggsndebs:
        pydist = pkg_resources.Distribution.from_filename(egginfo)
        try:
            dd.setdefault(
                pydist.project_name.lower(), {}).setdefault(
                pydist, set()).add(debname)
        except ValueError as le:
            log.warn("I got an error parsing a .egg-info file named \"%s\" "
                     "from Debian package \"%s\" as a pkg_resources "
                     "Distribution: %s" % (egginfo, debname, le,))
            pass

    # Now for each requirement, see if a Debian package satisfies it.
    ops = {'<': '<<', '>': '>>', '==': '=', '<=': '<=', '>=': '>='}
    for req in parsed_reqs:
        reqname = req.project_name.lower()
        gooddebs = set()
        for pydist, debs in dd.get(reqname, {}).iteritems():
            if pydist in req:
                # log.info("I found Debian packages \"%s\" which provides "
                #          "Python package \"%s\", version \"%s\", which "
                #          "satisfies our version requirements: \"%s\""
                #          % (', '.join(debs), req.project_name, ver, req)
                gooddebs |= (debs)
            else:
                log.info("I found Debian packages \"%s\" which provides "
                         "Python package \"%s\" which "
                         "does not satisfy our version requirements: "
                         "\"%s\" -- ignoring."
                         % (', '.join(debs), req.project_name, req))
        if not gooddebs:
            if on_failure == 'warn':
                log.warn(
                    "I found no Debian package which provides the required "
                    "Python package \"%s\" with version requirements "
                    "\"%s\"." % (req.project_name, req.specs))
            elif on_failure == "raise":
                raise CantSatisfyRequirement(
                    "I found no Debian package which provides the required "
                    "Python package \"%s\" with version requirements "
                    "\"%s\"." % (req.project_name, req.specs), req)
            elif on_failure == "guess":
                log.warn("I found no Debian package which provides the "
                         "required Python package \"%s\" with version "
                         "requirements \"%s\".  Guessing blindly that the "
                         "name \"python-%s\" will be it, and that the Python "
                         "package version number requirements will apply to "
                         "the Debian package." % (req.project_name,
                                                  req.specs, reqname))
                gooddebs.add("python-" + reqname)
        elif len(gooddebs) == 1:
            log.info("I found a Debian package which provides the require "
                     "Python package.  Python package: \"%s\", "
                     "Debian package: \"%s\";  adding Depends specifications "
                     "for the following version(s): \"%s\""
                     % (req.project_name, tuple(gooddebs)[0], req.specs))
        else:
            log.warn("I found multiple Debian packages which provide the "
                     "Python distribution required.  I'm listing them all "
                     "as alternates.  Candidate debs which claim to provide "
                     "the Python package \"%s\" are: \"%s\""
                     % (req.project_name, ', '.join(gooddebs),))

        alts = []
        for deb in gooddebs:
            added_any_alt = False
            for spec in req.specs:
                # Here we blithely assume that the Debian package
                # versions are enough like the Python package versions
                # that the requirement can be ported straight over...
                alts.append("%s (%s %s)" % (deb, ops[spec[0]], spec[1]))
                added_any_alt = True

            if not added_any_alt:
                # No alternates were added, but we have the name of a
                # good package.
                alts.append("%s" % deb)

        if len(alts):
            depends.append(' | '.join(alts))

    return depends


def make_tarball(tarball_fname, directory, cwd=None):
    "create a tarball from a directory"
    if tarball_fname.endswith('.gz'):
        opts = 'czf'
    else:
        opts = 'cf'
    args = ['/bin/tar', opts, tarball_fname, directory]
    process_command(args, cwd=cwd)


def expand_tarball(tarball_fname, cwd=None):
    "expand a tarball"
    if tarball_fname.endswith('.gz'):
        opts = 'xzf'
    elif tarball_fname.endswith('.bz2'):
        opts = 'xjf'
    else:
        opts = 'xf'
    args = ['/bin/tar', opts, tarball_fname]
    process_command(args, cwd=cwd)


def expand_zip(zip_fname, cwd=None):
    "expand a zip"
    unzip_path = '/usr/bin/unzip'
    if not os.path.exists(unzip_path):
        log.error('ERROR: {} does not exist'.format(unzip_path))
        sys.exit(1)
    args = [unzip_path, zip_fname]
    # Does it have a top dir
    res = subprocess.Popen(
        [args[0], '-l', args[1]], cwd=cwd,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    contents = []
    for line in res.stdout.readlines()[3:-2]:
        contents.append(line.split(None, 3)[-1])
    commonprefix = os.path.commonprefix(contents)
    if not commonprefix:
        extdir = os.path.join(cwd, os.path.basename(zip_fname[:-4]))
        args.extend(['-d', os.path.abspath(extdir)])

    process_command(args, cwd=cwd)


def expand_sdist_file(sdist_file, cwd=None):
    lower_sdist_file = sdist_file.lower()
    if lower_sdist_file.endswith('.zip'):
        expand_zip(sdist_file, cwd=cwd)
    elif lower_sdist_file.endswith('.tar.bz2'):
        expand_tarball(sdist_file, cwd=cwd)
    elif lower_sdist_file.endswith('.tar.gz'):
        expand_tarball(sdist_file, cwd=cwd)
    else:
        raise RuntimeError('could not guess format of original sdist file')


def repack_tarball_with_debianized_dirname(orig_sdist_file,
                                           repacked_sdist_file,
                                           debianized_dirname,
                                           original_dirname):
    working_dir = tempfile.mkdtemp()
    expand_sdist_file(orig_sdist_file, cwd=working_dir)
    fullpath_original_dirname = os.path.join(working_dir, original_dirname)
    fullpath_debianized_dirname = os.path.join(working_dir, debianized_dirname)

    # ensure sdist looks like sdist:
    assert os.path.exists(fullpath_original_dirname)
    assert len(os.listdir(working_dir)) == 1

    if fullpath_original_dirname != fullpath_debianized_dirname:
        # rename original dirname to debianized dirname
        os.rename(fullpath_original_dirname,
                  fullpath_debianized_dirname)
    make_tarball(repacked_sdist_file, debianized_dirname, cwd=working_dir)
    shutil.rmtree(working_dir)


def dpkg_buildpackage(*args, **kwargs):
    cwd = kwargs.pop('cwd', None)
    if len(kwargs) != 0:
        raise ValueError('only kwarg can be "cwd"')
    "call dpkg-buildpackage [arg1] [...] [argN]"
    args = ['/usr/bin/dpkg-buildpackage']+list(args)
    process_command(args, cwd=cwd)


def dpkg_source(b_or_x, arg1, cwd=None):
    "call dpkg-source -b|x arg1 [arg2]"
    assert b_or_x in ['-b', '-x']
    args = ['/usr/bin/dpkg-source', b_or_x, arg1]
    process_command(args, cwd=cwd)


def apply_patch(patchfile, cwd=None, posix=False, level=0):
    """call 'patch -p[level] [--posix] < arg1'

    posix mode is sometimes necessary. It keeps empty files so that
    dpkg-source removes their contents.

    """
    if not os.path.exists(patchfile):
        raise RuntimeError('patchfile "%s" does not exist' % patchfile)
    fd = open(patchfile, mode='r')

    level_str = '-p%d' % level
    args = ['/usr/bin/patch', level_str]
    if posix:
        args.append('--posix')

    log.info('PATCH COMMAND: %s < %s', ' '.join(args), patchfile)
    log.info('  PATCHING in dir: %s', cwd)
#    print >> sys.stderr, 'PATCH COMMAND:',' '.join(args),'<',patchfile
#    print >> sys.stderr, '  PATCHING in dir:',cwd
    res = subprocess.Popen(
        args, cwd=cwd,
        stdin=fd,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        universal_newlines=True
        )
    returncode = None
    while returncode is None:
        returncode = res.poll()
        ready = select.select([res.stdout, res.stderr], [], [], 0.1)
        # XXX figure out how to do this without reading byte-by-byte
        if res.stdout in ready[0]:
            sys.stdout.write(res.stdout.read(1))
            sys.stdout.flush()
        if res.stderr in ready[0]:
            sys.stderr.write(res.stderr.read(1))
            sys.stderr.flush()
    # finish outputting file
    sys.stdout.write(res.stdout.read())
    sys.stdout.flush()
    sys.stderr.write(res.stderr.read())
    sys.stderr.flush()

    if returncode:
        log.error('ERROR running: %s', ' '.join(args))
        log.error('ERROR in %s', cwd)
#        print >> sys.stderr, 'ERROR running: %s'%(' '.join(args),)
#        print >> sys.stderr, 'ERROR in',cwd
        raise RuntimeError('returncode %d' % returncode)


def parse_vals(cfg, section, option):
    """parse comma separated values in debian control file style from .cfg"""
    try:
        vals = cfg.get(section, option)
    except ConfigParser.NoSectionError as err:
        if section != 'DEFAULT':
            vals = cfg.get('DEFAULT', option)
        else:
            raise err
    vals = vals.split('#')[0]
    vals = vals.strip()
    vals = vals.split(',')
    vals = [v.strip() for v in vals]
    vals = [v for v in vals if len(v)]
    return vals


def parse_val(cfg, section, option):
    """extract a single value from .cfg"""
    vals = parse_vals(cfg, section, option)
    if len(vals) == 0:
        return ''
    else:
        assert len(vals) == 1, (section, option, vals, type(vals))
    return vals[0]


def apt_cache_info(apt_cache_cmd, package_name):
    if apt_cache_cmd not in ('showsrc', 'show'):
        raise NotImplementedError(
            "don't know how to run apt-cache command '%s'" % apt_cache_cmd)

    result_list = []
    args = ["apt-cache", apt_cache_cmd, package_name]
    cmd = subprocess.Popen(args,
                           stdin=subprocess.PIPE,
                           stderr=subprocess.PIPE,
                           stdout=subprocess.PIPE)
    returncode = cmd.wait()
    if returncode:
        errline = cmd.stderr.read()
        if not (
            returncode == 100 and errline ==
            "E: You must put some 'source' URIs in your sources.list\n"
        ):
            log.error('ERROR running: %s', ' '.join(args))
            raise RuntimeError(
                'returncode %d from subprocess %s' % (returncode, args))
    inlines = cmd.stdout.read()
    version_blocks = inlines.split('\n\n')
    for version_block in version_blocks:
        block_dict = {}

        if len(version_block) == 0:
            continue
        version_lines = version_block.split('\n')
        assert version_lines[0].startswith('Package: ')
        block_dict['Package'] = version_lines[0][len('Package: '):]

        if apt_cache_cmd == 'showsrc':
            assert version_lines[1].startswith('Binary: ')
            block_dict['Binary'] = version_lines[1][len('Binary: '):]
            block_dict['Binary'] = block_dict['Binary'].split(', ')

        elif apt_cache_cmd == 'show':
            for start in ('Provides: ', 'Conflicts: ', 'Replaces: '):
                key = start[:-2]
                for line in version_lines[2:]:
                    if line.startswith(start):
                        unsplit_line_result = line[len(start):]
                        split_result = unsplit_line_result.split(', ')
                        block_dict[key] = split_result
                if key not in block_dict:
                    block_dict[key] = []
        result_list.append(block_dict)
    return result_list


def check_cfg_files(cfg_files, module_name):
    """check if the configuration files actually specify something

    If config files are given, give warning if they don't contain
    information. This may indicate a wrong module name name, for
    example.
    """

    cfg = ConfigParser.ConfigParser()
    cfg.read(cfg_files)
    if cfg.has_section(module_name):
        section_items = cfg.items(module_name)
    else:
        section_items = []
    default_items = cfg.items('DEFAULT')

    n_items = len(section_items) + len(default_items)
    if n_items == 0:
        log.warn('configuration files were specified, but no options were '
                 'found in "%s" or "DEFAULT" sections.' % (module_name,))


class DebianInfo:
    """encapsulate information for Debian distribution system"""
    def __init__(self,
                 cfg_files=NotGiven,
                 module_name=NotGiven,
                 default_distribution=NotGiven,
                 guess_maintainer=NotGiven,
                 upstream_version=NotGiven,
                 has_ext_modules=NotGiven,
                 description=NotGiven,
                 long_description=NotGiven,
                 homepage=NotGiven,
                 patch_file=None,
                 patch_level=None,
                 setup_requires=None,
                 debian_version=None,
                 use_setuptools=False,
                 guess_conflicts_provides_replaces=False,
                 sdist_dsc_command=None,
                 with_python2=None,
                 with_python3=None,
                 no_python2_scripts=None,
                 no_python3_scripts=None,
                 force_x_python3_version=False,
                 allow_virtualenv_install_location=False,
                 compat=DH_DEFAULT_VERS,
                 with_dh_virtualenv=False,
                 with_dh_systemd=False,
                 ):
        if cfg_files is NotGiven:
            raise ValueError("cfg_files must be supplied")
        if module_name is NotGiven:
            raise ValueError("module_name must be supplied")
        if default_distribution is NotGiven:
            raise ValueError("default_distribution must be supplied")
        if guess_maintainer is NotGiven:
            raise ValueError("guess_maintainer must be supplied")
        if upstream_version is NotGiven:
            raise ValueError("upstream_version must be supplied")
        if has_ext_modules is NotGiven:
            raise ValueError("has_ext_modules must be supplied")
        if description is NotGiven:
            raise ValueError("description must be supplied")
        if long_description is NotGiven:
            raise ValueError("long_description must be supplied")
        if homepage is NotGiven:
            raise ValueError("homepage must be supplied")

        cfg_defaults = self._make_cfg_defaults(
            module_name=module_name,
            default_distribution=default_distribution,
            guess_maintainer=guess_maintainer,
            )

        if len(cfg_files):
            check_cfg_files(cfg_files, module_name)

        cfg = ConfigParser.ConfigParser(cfg_defaults)
        for cfg_file in cfg_files:
            with codecs.open(cfg_file, mode='r', encoding='utf-8') as fd:
                cfg.read_file(fd)

        if sdist_dsc_command is not None:
            # Allow distutils commands to override config files (this lets
            # command line options beat file options).
            for longopt, shortopt, desc in stdeb_cfg_options:
                opt_name = longopt[:-1]
                name = opt_name.replace('-', '_')
                value = getattr(sdist_dsc_command, name)
                if value is not None:
                    if not cfg.has_section(module_name):
                        cfg.add_section(module_name)
                    cfg.set(module_name, opt_name, value)

        self.stdeb_version = __stdeb_version__
        self.module_name = module_name
        self.compat = compat
        self.source = parse_val(cfg, module_name, 'Source')
        self.package = parse_val(cfg, module_name, 'Package')
        self.package3 = parse_val(cfg, module_name, 'Package3')
        forced_upstream_version = parse_val(cfg, module_name,
                                            'Forced-Upstream-Version')
        if forced_upstream_version == '':
            upstream_version_prefix = parse_val(cfg, module_name,
                                                'Upstream-Version-Prefix')
            upstream_version_suffix = parse_val(cfg, module_name,
                                                'Upstream-Version-Suffix')
            self.upstream_version = (upstream_version_prefix +
                                     debianize_version(upstream_version) +
                                     upstream_version_suffix)
        else:
            if (
                debianize_version(forced_upstream_version) !=
                forced_upstream_version
            ):
                raise ValueError(
                    'forced upstream version ("%s") not a '
                    'Debian-compatible version (e.g. "%s")' % (
                        forced_upstream_version,
                        debianize_version(forced_upstream_version)))
            self.upstream_version = forced_upstream_version
        self.epoch = parse_val(cfg, module_name, 'Epoch')
        if self.epoch != '' and not self.epoch.endswith(':'):
            self.epoch = self.epoch + ':'
        self.packaging_version = parse_val(cfg, module_name, 'Debian-Version')
        if debian_version is not None:
            # command-line arg overrides file
            self.packaging_version = debian_version
        self.dsc_version = '%s-%s' % (
            self.upstream_version,
            self.packaging_version)
        self.full_version = '%s%s-%s' % (
            self.epoch,
            self.upstream_version,
            self.packaging_version)
        self.distname = parse_val(cfg, module_name, 'Suite')
        self.distname3 = parse_val(cfg, module_name, 'Suite3')
        self.maintainer = ', '.join(parse_vals(cfg, module_name, 'Maintainer'))
        self.uploaders = parse_vals(cfg, module_name, 'Uploaders')
        self.date822 = get_date_822()

        build_deps = []
        if use_setuptools:
            if with_python2:
                build_deps.append('python-setuptools (>= 0.6b3)')
            if with_python3:
                build_deps.append('python3-setuptools')
        if setup_requires is not None and len(setup_requires):
            build_deps.extend(
                get_deb_depends_from_setuptools_requires(setup_requires))

        depends = ['${misc:Depends}', '${python:Depends}']
        depends3 = ['${misc:Depends}', '${python3:Depends}']
        need_custom_binary_target = False

        if has_ext_modules:
            self.architecture = 'any'
            if with_python2:
                build_deps.append(
                    'python-all-dev (>= %s)' % PYTHON_ALL_MIN_VERS)
            depends.append('${shlibs:Depends}')

            self.architecture3 = 'any'
            if with_python3:
                build_deps.append('python3-all-dev')
            depends3.append('${shlibs:Depends}')

        else:
            self.architecture = 'all'
            if with_python2:
                build_deps.append('python-all (>= %s)' % PYTHON_ALL_MIN_VERS)

            self.architecture3 = 'all'
            if with_python3:
                build_deps.append('python3-all')

        self.copyright_file = parse_val(cfg, module_name, 'Copyright-File')
        self.mime_file = parse_val(cfg, module_name, 'MIME-File')

        self.shared_mime_file = parse_val(cfg, module_name, 'Shared-MIME-File')

        if self.mime_file == '' and self.shared_mime_file == '':
            self.dh_installmime_indep_line = ''
        else:
            need_custom_binary_target = True
            self.dh_installmime_indep_line = '\tdh_installmime'

        mime_desktop_files = parse_vals(cfg, module_name, 'MIME-Desktop-Files')
        if len(mime_desktop_files):
            need_custom_binary_target = True

        #    E. any mime .desktop files
        self.install_file_lines = []
        for mime_desktop_file in mime_desktop_files:
            self.install_file_lines.append(
                '%s usr/share/applications' % mime_desktop_file)

        depends.extend(parse_vals(cfg, module_name, 'Depends'))
        depends3.extend(parse_vals(cfg, module_name, 'Depends3'))
        self.depends = ', '.join(depends)
        self.depends3 = ', '.join(depends3)

        self.debian_section = parse_val(cfg, module_name, 'Section')

        self.description = re.sub(r'\s+', ' ', description).strip()
        if long_description != 'UNKNOWN':
            ld2 = []
            for line in long_description.split('\n'):
                ls = line.strip()
                if len(ls):
                    ld2.append(' '+line)
                else:
                    ld2.append(' .')
            ld2 = ld2[:20]
            self.long_description = '\n'.join(ld2)
        else:
            self.long_description = ''

        build_deps.append('debhelper (>= %s)' % DH_MIN_VERS)

        build_deps.extend(parse_vals(cfg, module_name, 'Build-Depends'))
        self.build_depends = ', '.join(build_deps)

        suggests = ', '.join(parse_vals(cfg, module_name, 'Suggests'))
        suggests3 = ', '.join(parse_vals(cfg, module_name, 'Suggests3'))
        recommends = ', '.join(parse_vals(cfg, module_name, 'Recommends'))
        recommends3 = ', '.join(parse_vals(cfg, module_name, 'Recommends3'))

        self.source_stanza_extras = ''

        if homepage != 'UNKNOWN':
            self.source_stanza_extras += 'Homepage: %s\n' % homepage

        build_conflicts = parse_vals(cfg, module_name, 'Build-Conflicts')
        if len(build_conflicts):
            self.source_stanza_extras += ('Build-Conflicts: ' +
                                          ', '.join(build_conflicts)+'\n')

        self.patch_file = parse_val(cfg, module_name, 'Stdeb-Patch-File')

        if patch_file is not None:
            if self.patch_file != '':
                raise RuntimeError('A patch file was specified on the command '
                                   'line and in .cfg file.')
            else:
                self.patch_file = patch_file

        self.patch_level = parse_val(cfg, module_name, 'Stdeb-Patch-Level')
        if self.patch_level != '':
            if patch_level is not None:
                raise RuntimeError('A patch level was specified on the '
                                   'command line and in .cfg file.')
            else:
                self.patch_level = int(self.patch_level)
        else:
            if patch_level is not None:
                self.patch_level = patch_level
            else:
                self.patch_level = 0

        python2_depends_name = parse_val(
            cfg, module_name, 'Python2-Depends-Name')

        xs_python_version = parse_vals(cfg, module_name, 'XS-Python-Version')

        if len(xs_python_version) != 0:
            self.source_stanza_extras += ('X-Python-Version: ' +
                                          ', '.join(xs_python_version)+'\n')

        x_python3_version = parse_vals(cfg, module_name, 'X-Python3-Version')
        x_python3_version = [v.strip('"') for v in x_python3_version]

        if len(x_python3_version) != 0:
            self.source_stanza_extras += ('X-Python3-Version: ' +
                                          ', '.join(x_python3_version)+'\n')

        dpkg_shlibdeps_params = parse_val(
            cfg, module_name, 'dpkg-shlibdeps-params')
        if dpkg_shlibdeps_params:
            need_custom_binary_target = True
            self.dh_binary_arch_lines = """\tdh binary-arch --before dh_shlibdeps
\tdh_shlibdeps -a --dpkg-shlibdeps-params=%s
\tdh binary --after dh_shlibdeps""" % dpkg_shlibdeps_params
        else:
            self.dh_binary_arch_lines = '\tdh binary-arch'
        self.dh_binary_indep_lines = '\tdh binary-indep'

        dh_python3_params = parse_val(cfg, module_name, 'dh-python3-params')

        conflicts = parse_vals(cfg, module_name, 'Conflicts')
        conflicts3 = parse_vals(cfg, module_name, 'Conflicts3')
        breaks = parse_vals(cfg, module_name, 'Breaks')
        breaks3 = parse_vals(cfg, module_name, 'Breaks3')
        provides = parse_vals(cfg, module_name, 'Provides')
        provides3 = parse_vals(cfg, module_name, 'Provides3')
        replaces = parse_vals(cfg, module_name, 'Replaces')
        replaces3 = parse_vals(cfg, module_name, 'Replaces3')

        if guess_conflicts_provides_replaces:
            # Find list of binaries which we will conflict/provide/replace.

            cpr_binaries = set()

            # Get original Debian information for the package named the same.
            for version_info in apt_cache_info('showsrc', self.package):

                # Remember each of the binary packages produced by the Debian
                # source
                for binary in version_info['Binary']:
                    cpr_binaries.add(binary)

                # TODO: do this for every version available , just the
                # first, or ???
                break

            # Descend each of the original binaries and see what
            # packages they conflict/ provide/ replace:
            for orig_binary in cpr_binaries:
                for version_info in apt_cache_info('show', orig_binary):
                    provides.extend(version_info['Provides'])
                    provides3.extend(version_info['Provides'])
                    conflicts.extend(version_info['Conflicts'])
                    conflicts3.extend(version_info['Conflicts'])
                    replaces.extend(version_info['Replaces'])
                    replaces3.extend(version_info['Replaces'])

            if self.package in cpr_binaries:
                cpr_binaries.remove(self.package)  # don't include ourself

            cpr_binaries = list(cpr_binaries)  # convert to list

            conflicts.extend(cpr_binaries)
            conflicts3.extend(cpr_binaries)
            provides.extend(cpr_binaries)
            provides3.extend(cpr_binaries)
            replaces.extend(cpr_binaries)
            replaces3.extend(cpr_binaries)

            # round-trip through set to get unique entries
            conflicts = list(set(conflicts))
            conflicts3 = list(set(conflicts3))
            provides = list(set(provides))
            provides3 = list(set(provides3))
            replaces = list(set(replaces))
            replaces3 = list(set(replaces3))

        self.package_stanza_extras = ''
        self.package_stanza_extras3 = ''

        if len(conflicts):
            self.package_stanza_extras += (
                'Conflicts: ' + ', '.join(conflicts)+'\n')
        if len(conflicts3):
            self.package_stanza_extras3 += (
                'Conflicts: ' + ', '.join(conflicts3)+'\n')

        if len(breaks):
            self.package_stanza_extras += (
                'Breaks: ' + ', '.join(breaks)+'\n')
        if len(breaks3):
            self.package_stanza_extras3 += (
                'Breaks: ' + ', '.join(breaks3)+'\n')
        if len(provides):
            self.package_stanza_extras += (
                'Provides: ' + ', '.join(provides)+'\n')

        if len(provides3):
            self.package_stanza_extras3 += (
                'Provides: ' + ', '.join(provides3)+'\n')

        if len(replaces):
            self.package_stanza_extras += (
                'Replaces: ' + ', '.join(replaces)+'\n')
        if len(replaces3):
            self.package_stanza_extras3 += (
                'Replaces: ' + ', '.join(replaces3)+'\n')
        if len(recommends):
            self.package_stanza_extras += ('Recommends: ' + recommends+'\n')

        if len(recommends3):
            self.package_stanza_extras3 += ('Recommends: ' + recommends3+'\n')

        if len(suggests):
            self.package_stanza_extras += ('Suggests: ' + suggests+'\n')

        if len(suggests3):
            self.package_stanza_extras3 += ('Suggests: ' + suggests3+'\n')

        self.dirlist = ""

        if not (with_python2 or with_python3):
            raise RuntimeError('nothing to do - neither Python 2 or 3.')

        if with_python2:
            if shutil.which("python"):
                self.python2_binname = "python"
            elif shutil.which("python2"):
                self.python2_binname = "python2"
            else:
                raise RuntimeError("Python 2 binary not found on path as either `python` or `python2`")
        sequencer_with = []
        if with_python2:
            sequencer_with.append('python2')
        if with_python3:
            sequencer_with.append('python3')

        no_script_lines = []

        if no_python2_scripts:
            # install to a location where debian tools do not find them
            self.no_python2_scripts_cli_args = '--install-scripts=/trash'
            no_script_lines.append(
                'rm -rf debian/%s/trash' % (self.package,))
        else:
            self.no_python2_scripts_cli_args = ''
        if no_python3_scripts:
            # install to a location where debian tools do not find them
            self.no_python3_scripts_cli_args = '--install-scripts=/trash'
            no_script_lines.append(
                'rm -rf debian/%s/trash' % (self.package3,))
        else:
            self.no_python3_scripts_cli_args = ''

        self.scripts_cleanup = '\n'.join(
            ['        ' + s for s in no_script_lines])

        if sys.prefix != '/usr':
            if not allow_virtualenv_install_location:
                # virtualenv will set distutils
                # --prefix=/path/to/virtualenv, but unless explicitly
                # requested, we want to install into /usr.
                self.install_prefix = '--prefix=/usr'
            else:
                self.install_prefix = '--prefix=%s' % sys.prefix
        else:
            self.install_prefix = ''

        rules_override_clean_target_pythons = []
        rules_override_build_target_pythons = []
        rules_override_install_target_pythons = []
        if with_python2:
            rules_override_clean_target_pythons.append(
                RULES_OVERRIDE_CLEAN_TARGET_PY2 % self.__dict__
                )
            rules_override_build_target_pythons.append(
                RULES_OVERRIDE_BUILD_TARGET_PY2 % self.__dict__
                )
            rules_override_install_target_pythons.append(
                RULES_OVERRIDE_INSTALL_TARGET_PY2 % self.__dict__
                )
        if with_python3:
            rules_override_clean_target_pythons.append(
                RULES_OVERRIDE_CLEAN_TARGET_PY3 % self.__dict__
                )
            rules_override_build_target_pythons.append(
                RULES_OVERRIDE_BUILD_TARGET_PY3 % self.__dict__
                )
            rules_override_install_target_pythons.append(
                RULES_OVERRIDE_INSTALL_TARGET_PY3 % self.__dict__
                )
        self.rules_override_clean_target_pythons = \
            '\n'.join(rules_override_clean_target_pythons)
        self.rules_override_build_target_pythons = \
            '\n'.join(rules_override_build_target_pythons)
        self.rules_override_install_target_pythons = \
            '\n'.join(rules_override_install_target_pythons)

        self.override_dh_auto_clean = \
            RULES_OVERRIDE_CLEAN_TARGET % self.__dict__
        self.override_dh_auto_build = \
            RULES_OVERRIDE_BUILD_TARGET % self.__dict__
        self.override_dh_auto_install = \
            RULES_OVERRIDE_INSTALL_TARGET % self.__dict__

        scripts = ''
        if with_python2 and python2_depends_name:
            scripts = (
                '        sed -i ' +
                r'"s/\([ =]\)python[0-9]\?\(\(:any\)\? (\)/\\1%s\\2/g" ' +
                'debian/%s.substvars') % (python2_depends_name, self.package)
        self.override_dh_python2 = RULES_OVERRIDE_PYTHON2 % {
            'scripts': scripts
        }

        scripts = ''
        if force_x_python3_version and with_python3 and x_python3_version and \
                x_python3_version[0]:
            # override dh_python3 target to modify the dependencies
            # to ensure that the passed minimum X-Python3-Version is usedby
            version = x_python3_version[0]
            if not version.endswith('~'):
                version += '~'
                scripts = (
                    '        sed -i ' +
                    r'"s/\([ =]python3:any (\)>= [^)]*\()\)/\\1%s\\2/g" ' +
                    'debian/%s.substvars') % (version, self.package3)
        if scripts or dh_python3_params:
            self.override_dh_python3 = RULES_OVERRIDE_PYTHON3 % {
                'scripts': scripts,
                'dh_python3_params': dh_python3_params,
            }
        else:
            self.override_dh_python3 = ''

        sequencer_options = ['--with '+','.join(sequencer_with)]

        if with_dh_virtualenv:
            if with_python2:
                self.override_dh_virtualenv_py = \
                    RULES_OVERRIDE_DH_VIRTUALENV_PY2 % self.__dict__
            if with_python3:
                self.override_dh_virtualenv_py = \
                    RULES_OVERRIDE_DH_VIRTUALENV_PY3 % self.__dict__

            sequencer_options.append('--with python-virtualenv')
        else:
            sequencer_options.append('--buildsystem=pybuild')
            self.override_dh_virtualenv_py = ''

        if with_dh_systemd:
            sequencer_options.append('--with systemd')

        self.sequencer_options = ' '.join(sequencer_options)

        setup_env_vars = parse_vals(cfg, module_name, 'Setup-Env-Vars')
        self.exports = ""
        if len(setup_env_vars):
            self.exports += '\n'
            self.exports += '#exports specified using stdeb Setup-Env-Vars:\n'
            self.exports += '\n'.join(
                ['export %s' % v for v in setup_env_vars])
            self.exports += '\n'
        self.udev_rules = parse_val(cfg, module_name, 'Udev-Rules')

        if need_custom_binary_target:
            if self.architecture == 'all':
                self.binary_target_lines = (
                    RULES_BINARY_ALL_TARGET % self.__dict__ +
                    RULES_BINARY_INDEP_TARGET % self.__dict__)
            else:
                self.binary_target_lines = (
                    RULES_BINARY_TARGET % self.__dict__ +
                    RULES_BINARY_INDEP_TARGET % self.__dict__ +
                    RULES_BINARY_ARCH_TARGET % self.__dict__)
        else:
            self.binary_target_lines = ''

        self.changelog_distname = CHANGELOG_PY2_DISTNAME % self.__dict__
        if with_python2:
            self.control_py2_stanza = CONTROL_PY2_STANZA % self.__dict__
        else:
            self.control_py2_stanza = ''

        if with_python3:
            self.control_py3_stanza = CONTROL_PY3_STANZA % self.__dict__
            if self.distname3 and self.distname3 != self.distname:
                if with_python2:
                    raise ValueError(
                        "Suites are shared between versions. To use a "
                        "different value for Suite3 run --with-python3 true "
                        "--with-python2 false only.")
                self.changelog_distname = \
                    CHANGELOG_PY3_DISTNAME % self.__dict__
        else:
            self.control_py3_stanza = ''

        self.with_python2 = with_python2
        self.with_python3 = with_python3
        self.no_python2_scripts = no_python2_scripts
        self.no_python3_scripts = no_python3_scripts

    def _make_cfg_defaults(self,
                           module_name=NotGiven,
                           default_distribution=NotGiven,
                           guess_maintainer=NotGiven,
                           ):
        defaults = {}
        default_re = re.compile(r'^.* \(Default: (.*)\)$')
        for longopt, shortopt, description in stdeb_cfg_options:
            assert longopt.endswith('=')
            assert longopt.lower() == longopt
            key = longopt[:-1]
            matchobj = default_re.search(description)
            if matchobj is not None:
                # has a default value
                groups = matchobj.groups()
                assert len(groups) == 1
                value = groups[0]
                # A few special cases
                if value == '<source-debianized-setup-name>':
                    assert key == 'source'
                    value = source_debianize_name(module_name)
                elif value == 'python-<debianized-setup-name>':
                    assert key == 'package'
                    value = 'python-' + debianize_name(module_name)
                elif value == 'python3-<debianized-setup-name>':
                    assert key == 'package3'
                    value = 'python3-' + debianize_name(module_name)
                elif value == '<setup-maintainer-or-author>':
                    assert key == 'maintainer'
                    value = guess_maintainer
                elif value == '<same-as-suite>':
                    assert key == 'suite3'
                    # Set to empty string so value of suite is used.
                    value = ''
                if key == 'suite':
                    if default_distribution is not None:
                        value = default_distribution
                        log.warn('Deprecation warning: you are using the '
                                 '--default-distribution option. '
                                 'Switch to the --suite option.')
            else:
                # no default value
                value = ''
            defaults[key] = value
        return defaults


def build_dsc(debinfo,
              dist_dir,
              repackaged_dirname,
              orig_sdist=None,
              patch_posix=0,
              remove_expanded_source_dir=0,
              debian_dir_only=False,
              sign_dsc=False,
              sign_key=None,
              ignore_source_changes=False,
              ):
    """make debian source package"""
    #    A. Find new dirname and delete any pre-existing contents

    # dist_dir is usually 'deb_dist'

    # the location of the copied original source package (it was
    # re-recreated in dist_dir)
    if debian_dir_only:
        fullpath_repackaged_dirname = os.path.abspath(os.curdir)
    else:
        fullpath_repackaged_dirname = os.path.join(
            dist_dir, repackaged_dirname)

    ###############################################
    # 1. make temporary original source tarball

    #    Note that, for the final tarball, best practices suggest
    #    using "dpkg-source -b".  See
    #    http://www.debian.org/doc/developers-reference/ch-best-pkging-practices.en.html

    # Create the name of the tarball that qualifies as the upstream
    # source. If the original was specified, we'll link to
    # it. Otherwise, we generate our own .tar.gz file from the output
    # of "python setup.py sdist" (done above) so that we avoid
    # packaging .svn directories, for example.

    if not debian_dir_only:
        repackaged_orig_tarball = \
            '%(source)s_%(upstream_version)s.orig.tar.gz' % debinfo.__dict__
        repackaged_orig_tarball_path = os.path.join(dist_dir,
                                                    repackaged_orig_tarball)
        if orig_sdist is not None:
            if os.path.exists(repackaged_orig_tarball_path):
                os.unlink(repackaged_orig_tarball_path)
            link_func(orig_sdist, repackaged_orig_tarball_path)
        else:
            make_tarball(repackaged_orig_tarball,
                         repackaged_dirname,
                         cwd=dist_dir)

        # apply patch
        if debinfo.patch_file != '':
            apply_patch(debinfo.patch_file,
                        posix=patch_posix,
                        level=debinfo.patch_level,
                        cwd=fullpath_repackaged_dirname)

    for fname in ['Makefile', 'makefile']:
        if os.path.exists(os.path.join(fullpath_repackaged_dirname, fname)):
            sys.stderr.write('*'*1000 + '\n')
            sys.stderr.write('WARNING: a Makefile exists in this package. '
                             'stdeb will tell debhelper 7 to use setup.py '
                             'to build and install the package, and the '
                             'Makefile will be ignored.\n')
            sys.stderr.write('*'*1000 + '\n')

    ###############################################
    # 2. create debian/ directory and contents
    debian_dir = os.path.join(fullpath_repackaged_dirname, 'debian')
    if not os.path.exists(debian_dir):
        os.mkdir(debian_dir)

    #    A. debian/changelog
    changelog = CHANGELOG_FILE % debinfo.__dict__
    with codecs.open(os.path.join(debian_dir, 'changelog'),
                     mode='w', encoding='utf-8') as fd:
        fd.write(changelog)

    #    B. debian/control
    if debinfo.uploaders:
        debinfo.uploaders = 'Uploaders: %s\n' % ', '.join(debinfo.uploaders)
    else:
        debinfo.uploaders = ''
    control = CONTROL_FILE % debinfo.__dict__
    control = re.sub('\n{3,}', '\n\n', control)
    with codecs.open(os.path.join(debian_dir, 'control'),
                     mode='w', encoding='utf-8') as fd:
        fd.write(control)

    #    C. debian/rules
    debinfo.percent_symbol = '%'
    rules = RULES_MAIN % debinfo.__dict__

    rules = rules.replace('        ', '\t')
    rules = re.sub('\n{3,}', '\n\n', rules)
    rules_fname = os.path.join(debian_dir, 'rules')
    with codecs.open(rules_fname,
                     mode='w', encoding='utf-8') as fd:
        fd.write(rules)
    os.chmod(rules_fname, 0o755)

    #    D. debian/compat
    fd = open(os.path.join(debian_dir, 'compat'), mode='w')
    fd.write('{}\n'.format(str(debinfo.compat)))
    fd.close()

    #    E. debian/package.mime
    if debinfo.mime_file != '':
        if not os.path.exists(debinfo.mime_file):
            raise ValueError(
                'a MIME file was specified, but does not exist: %s' % (
                    debinfo.mime_file,))
        link_func(debinfo.mime_file,
                  os.path.join(debian_dir, debinfo.package + '.mime'))
    if debinfo.shared_mime_file != '':
        if not os.path.exists(debinfo.shared_mime_file):
            raise ValueError(
                'a shared MIME file was specified, but does not exist: %s' % (
                    debinfo.shared_mime_file,))
        link_func(debinfo.shared_mime_file,
                  os.path.join(
                      debian_dir, debinfo.package + '.sharedmimeinfo'))

    #    F. debian/copyright
    if debinfo.copyright_file != '':
        link_func(debinfo.copyright_file,
                  os.path.join(debian_dir, 'copyright'))

    #    H. debian/<package>.install
    if len(debinfo.install_file_lines):
        fd = open(
            os.path.join(debian_dir, '%s.install' % debinfo.package), mode='w')
        fd.write('\n'.join(debinfo.install_file_lines)+'\n')
        fd.close()

    #    I. debian/<package>.udev
    if debinfo.udev_rules != '':
        fname = debinfo.udev_rules
        if not os.path.exists(fname):
            raise ValueError('udev rules file specified, but does not exist')
        if debinfo.with_python2:
            link_func(fname,
                      os.path.join(debian_dir, '%s.udev' % debinfo.package))
        if debinfo.with_python3:
            link_func(fname,
                      os.path.join(debian_dir, '%s.udev' % debinfo.package3))

    #    J. debian/source/format
    os.mkdir(os.path.join(debian_dir, 'source'))
    fd = open(os.path.join(debian_dir, 'source', 'format'), mode='w')
    fd.write('3.0 (quilt)\n')
    fd.close()

    fd = open(os.path.join(debian_dir, 'source', 'options'), mode='w')
    fd.write(r'extend-diff-ignore="\.egg-info$"')
    fd.close()

    if debian_dir_only:
        return

    ###############################################
    # 3. unpack original source tarball

    debianized_package_dirname = fullpath_repackaged_dirname+'.debianized'
    if os.path.exists(debianized_package_dirname):
        raise RuntimeError('debianized_package_dirname exists: %s' %
                           debianized_package_dirname)
    #    A. move debianized tree away
    os.rename(fullpath_repackaged_dirname, debianized_package_dirname)
    if orig_sdist is not None:
        #    B. expand repackaged original tarball
        tmp_dir = os.path.join(dist_dir, 'tmp-expand')
        os.mkdir(tmp_dir)
        try:
            expand_tarball(orig_sdist, cwd=tmp_dir)
            orig_tarball_top_contents = os.listdir(tmp_dir)

            # make sure original tarball has exactly one directory
            assert len(orig_tarball_top_contents) == 1
            orig_dirname = orig_tarball_top_contents[0]
            fullpath_orig_dirname = os.path.join(tmp_dir, orig_dirname)

            #    C. remove original repackaged tree
            shutil.rmtree(fullpath_orig_dirname)

        finally:
            shutil.rmtree(tmp_dir)

    if 1:
        # check versions of debhelper and python-all
        debhelper_version_str = get_version_str('debhelper')
        if len(debhelper_version_str) == 0:
            log.warn('This version of stdeb requires debhelper >= %s, but you '
                     'do not have debhelper installed. '
                     'Could not check compatibility.' % DH_MIN_VERS)
        else:
            if not dpkg_compare_versions(
                debhelper_version_str, 'ge', DH_MIN_VERS
            ):
                log.warn('This version of stdeb requires debhelper >= %s. '
                         'Use stdeb 0.9.x to generate source packages '
                         'compatible with older versions of debhelper.' % (
                            DH_MIN_VERS,))

        if debinfo.with_python2:
            python_defaults_version_str = get_version_str('python-all')

            if len(python_defaults_version_str) == 0:
                log.warn('This version of stdeb requires python-all >= %s, '
                         'but you do not have this package installed. Could '
                         'not check compatibility.' % PYTHON_ALL_MIN_VERS)
            else:
                if not dpkg_compare_versions(
                    python_defaults_version_str, 'ge', PYTHON_ALL_MIN_VERS
                ):
                    log.warn('This version of stdeb requires python-all >= '
                             '%s. Use stdeb 0.6.0 or older to generate source '
                             'packages that use python-support.' % (
                                 PYTHON_ALL_MIN_VERS,))

        if debinfo.with_python3:
            python3_defaults_version_str = get_version_str('python3-all')

            if len(python3_defaults_version_str) == 0:
                log.warn('This version of stdeb requires python3-all, '
                         'but you do not have this package installed.')
            else:
                if not dpkg_compare_versions(
                    python3_defaults_version_str, 'ge', PYTHON3_ALL_MIN_VERS
                ):
                    log.warn('This version of stdeb requires python-all >= '
                             '%s. Use stdeb 0.6.0 or older to generate source '
                             'packages that use python-support.' % (
                                 PYTHON_ALL_MIN_VERS,))

    #    D. restore debianized tree
    os.rename(fullpath_repackaged_dirname+'.debianized',
              fullpath_repackaged_dirname)

    #    Re-generate tarball using best practices see
    #    http://www.debian.org/doc/developers-reference/ch-best-pkging-practices.en.html

    args = ['-S', '-sa']

    if not sign_dsc:
        args += ['-uc', '-us']
    elif sign_key is not None:
        args += ['--sign-key={}'.format(sign_key)]

    if ignore_source_changes:
        args.append('-i.*')

    dpkg_buildpackage(*args, cwd=fullpath_repackaged_dirname)

    if 1:
        shutil.rmtree(fullpath_repackaged_dirname)

    if not remove_expanded_source_dir:
        # expand the debian source package
        dsc_name = debinfo.source + '_' + debinfo.dsc_version + '.dsc'
        dpkg_source('-x', dsc_name,
                    cwd=dist_dir)


CHANGELOG_PY2_DISTNAME = '%(distname)s'
CHANGELOG_PY3_DISTNAME = '%(distname3)s'
CHANGELOG_FILE = """\
%(source)s (%(full_version)s) %(changelog_distname)s; urgency=low

  * source package automatically created by stdeb %(stdeb_version)s

 -- %(maintainer)s  %(date822)s\n"""

CONTROL_FILE = """\
Source: %(source)s
Maintainer: %(maintainer)s
%(uploaders)sSection: %(debian_section)s
Priority: optional
Build-Depends: %(build_depends)s
Standards-Version: 4.7.0
%(source_stanza_extras)s

%(control_py2_stanza)s

%(control_py3_stanza)s
"""

CONTROL_PY2_STANZA = """
Package: %(package)s
Architecture: %(architecture)s
Depends: %(depends)s
%(package_stanza_extras)sDescription: %(description)s
%(long_description)s
"""

CONTROL_PY3_STANZA = """
Package: %(package3)s
Architecture: %(architecture3)s
Depends: %(depends3)s
%(package_stanza_extras3)sDescription: %(description)s
%(long_description)s
"""

RULES_MAIN = """\
#!/usr/bin/make -f

# This file was automatically generated by stdeb %(stdeb_version)s at
# %(date822)s
%(exports)s
%(percent_symbol)s:
        dh $@ %(sequencer_options)s

%(override_dh_auto_clean)s

%(override_dh_auto_build)s

%(override_dh_auto_install)s

%(override_dh_python2)s

%(override_dh_python3)s

%(override_dh_virtualenv_py)s

%(binary_target_lines)s
"""

RULES_OVERRIDE_CLEAN_TARGET_PY2 = "        %(python2_binname)s setup.py clean -a"
RULES_OVERRIDE_CLEAN_TARGET_PY3 = "        python3 setup.py clean -a"
RULES_OVERRIDE_CLEAN_TARGET = r"""
override_dh_auto_clean:
%(rules_override_clean_target_pythons)s
        find . -name \*.pyc -exec rm {} \;
"""

RULES_OVERRIDE_BUILD_TARGET_PY2 = "        %(python2_binname)s setup.py build --force"
RULES_OVERRIDE_BUILD_TARGET_PY3 = "        python3 setup.py build --force"
RULES_OVERRIDE_BUILD_TARGET = """
override_dh_auto_build:
%(rules_override_build_target_pythons)s
"""

RULES_OVERRIDE_INSTALL_TARGET_PY2 = "        %(python2_binname)s setup.py install --force --root=debian/%(package)s --no-compile -O0 --install-layout=deb %(install_prefix)s %(no_python2_scripts_cli_args)s"  # noqa: E501

RULES_OVERRIDE_INSTALL_TARGET_PY3 = "        python3 setup.py install --force --root=debian/%(package3)s --no-compile -O0 --install-layout=deb %(install_prefix)s %(no_python3_scripts_cli_args)s"  # noqa: E501

RULES_OVERRIDE_INSTALL_TARGET = """
override_dh_auto_install:
%(rules_override_install_target_pythons)s
%(scripts_cleanup)s
"""

RULES_OVERRIDE_PYTHON2 = """
override_dh_python2:
        dh_python2 --no-guessing-versions
%(scripts)s
"""
RULES_OVERRIDE_PYTHON3 = """
override_dh_python3:
        dh_python3 %(dh_python3_params)s
%(scripts)s
"""

RULES_OVERRIDE_DH_VIRTUALENV_PY2 = """
override_dh_virtualenv:
        dh_virtualenv --python python2
"""

RULES_OVERRIDE_DH_VIRTUALENV_PY3 = """
override_dh_virtualenv:
        dh_virtualenv --python python3
"""

RULES_BINARY_TARGET = """
binary: binary-arch binary-indep
"""

RULES_BINARY_ALL_TARGET = """
binary: binary-indep
"""

RULES_BINARY_ARCH_TARGET = """
binary-arch: build
%(dh_binary_arch_lines)s
"""

RULES_BINARY_INDEP_TARGET = """
binary-indep: build
%(dh_binary_indep_lines)s
%(dh_installmime_indep_line)s
"""



import re
import requests
import stdeb

USER_AGENT = 'pypi-install/%s ( https://github.com/astraw/stdeb )' % \
    stdeb.__version__


def normalize_package_name(package_name):
    return re.sub(r"[-_.]+", "-", package_name).lower()


class PyPIClient(object):
    def __init__(self, pypi_url="https://pypi.org", user_agent=USER_AGENT):
        self.pypi_url = pypi_url
        self._http_session = requests.Session()
        self._http_session.headers["User-Agent"] = user_agent

    def release_version(self, package_name):
        package_name = normalize_package_name(package_name)
        response = self._http_session.get("%s/pypi/%s/json" % (self.pypi_url, package_name))
        data = response.json()
        return data["info"]["version"]

    def release_versions(self, package_name):
        package_name = normalize_package_name(package_name)
        response = self._http_session.get(
            "%s/simple/%s" % (self.pypi_url, package_name),
            headers={"Accept": "application/vnd.pypi.simple.latest+json"})
        data = response.json()
        return data["versions"]

    def download_url(self, package_name, version):
        download_url = None
        md5_digest = None
        package_name = normalize_package_name(package_name)
        response = self._http_session.get(
            "%s/pypi/%s/%s/json" % (self.pypi_url, package_name, version))
        data = response.json()
        for url in data["urls"]:
            if url["packagetype"] == "sdist" and \
                    url["python_version"] == "source" and \
                    url["url"].endswith((".tar.gz", ".zip")):
                download_url = url["url"]
                md5_digest = url["digests"]["md5"]
                break

        return download_url, md5_digest



from __future__ import print_function
import os
from functools import partial
import requests
import hashlib
import warnings
from stdeb.pypi_simple import PyPIClient, USER_AGENT

myprint = print


def find_tar_gz(package_name, pypi_url='https://pypi.org',
                verbose=0, release=None):
    pypi_client = PyPIClient(pypi_url, USER_AGENT)

    download_url = None
    expected_md5_digest = None

    if verbose >= 2:
        myprint('querying PyPI (%s) for package name "%s"' % (pypi_url,
                                                              package_name))

    all_releases = pypi_client.release_versions(package_name)
    if release is not None:
        # A specific release is requested.
        if verbose >= 2:
            myprint(
                'found all available releases: %s' %
                (', '.join(all_releases),))

        if release not in all_releases:
            raise ValueError('your desired release %r is not among available '
                             'releases %r' % (release, all_releases))
        version = release
    else:
        default_release = pypi_client.release_version(package_name)
        if verbose >= 2:
            myprint(
                'found default release: %s' % (', '.join(default_release),))

        version = default_release

    download_url, expected_md5_digest = pypi_client.download_url(package_name, version)

    if download_url is None:
        raise ValueError('no package "%s" was found' % package_name)
    return download_url, expected_md5_digest


def md5sum(filename):
    # from
    # http://stackoverflow.com/questions/7829499/using-hashlib-to-compute-md5-digest-of-a-file-in-python-3
    with open(filename, mode='rb') as f:
        d = hashlib.md5()
        for buf in iter(partial(f.read, 128), b''):
            d.update(buf)
    return d.hexdigest()


def get_source_tarball(package_name, verbose=0, allow_unsafe_download=False,
                       release=None):
    download_url, expected_md5_digest = find_tar_gz(package_name,
                                                    verbose=verbose,
                                                    release=release)
    if not download_url.startswith('https://'):
        if allow_unsafe_download:
            warnings.warn('downloading from unsafe url: %r' % download_url)
        else:
            raise ValueError('PYPI returned unsafe url: %r' % download_url)

    fname = download_url.split('/')[-1]
    if expected_md5_digest is not None:
        if os.path.exists(fname):
            actual_md5_digest = md5sum(fname)
            if actual_md5_digest == expected_md5_digest:
                if verbose >= 1:
                    myprint('Download URL: %s' % download_url)
                    myprint(
                        'File "%s" already exists with correct checksum.' %
                        fname)
                return fname
            else:
                raise ValueError(
                    'File "%s" exists but has wrong checksum.' % fname)
    if verbose >= 1:
        myprint('downloading %s' % download_url)
    headers = {'User-Agent': USER_AGENT}
    r = requests.get(download_url, headers=headers)
    r.raise_for_status()
    package_tar_gz = r.content
    if verbose >= 1:
        myprint('done downloading %d bytes.' % (len(package_tar_gz), ))
    if expected_md5_digest is not None:
        m = hashlib.md5()
        m.update(package_tar_gz)
        actual_md5_digest = m.hexdigest()
        if verbose >= 2:
            myprint('md5:   actual %s\n     expected %s' %
                    (actual_md5_digest, expected_md5_digest))
        if actual_md5_digest != expected_md5_digest:
            raise ValueError('actual and expected md5 digests do not match')
    else:
        warnings.warn('no md5 digest found -- cannot verify source file')

    fd = open(fname, mode='wb')
    fd.write(package_tar_gz)
    fd.close()
    return fname



import os
import shutil
import tempfile

from stdeb import log
from stdeb.util import expand_sdist_file
from stdeb.util import build_dsc, stdeb_cmdline_opts, \
     stdeb_cmd_bool_opts, stdeb_cfg_options
from stdeb.util import repack_tarball_with_debianized_dirname
from stdeb.command.common import common_debian_package_command

__all__ = ['sdist_dsc']


class sdist_dsc(common_debian_package_command):
    description = "distutils command to create a debian source distribution"

    user_options = stdeb_cmdline_opts + [
        ('use-premade-distfile=', 'P',
         'use .zip or .tar.gz file already made by sdist command'),
        ] + stdeb_cfg_options

    boolean_options = stdeb_cmd_bool_opts

    def initialize_options(self):
        self.use_premade_distfile = None
        common_debian_package_command.initialize_options(self)

    def run(self):
        debinfo = self.get_debinfo()
        if debinfo.patch_file != '' and self.patch_already_applied:
            raise RuntimeError('A patch was already applied, but another '
                               'patch is requested.')

        repackaged_dirname = debinfo.source+'-'+debinfo.upstream_version
        fullpath_repackaged_dirname = os.path.join(self.dist_dir,
                                                   repackaged_dirname)

        cleanup_dirs = []
        if self.use_premade_distfile is None:
            # generate original tarball
            sdist_cmd = self.distribution.get_command_obj('sdist')
            self.run_command('sdist')

            source_tarball = None
            for archive_file in sdist_cmd.get_archive_files():
                if archive_file.endswith('.tar.gz'):
                    source_tarball = archive_file

            if source_tarball is None:
                raise RuntimeError('sdist did not produce .tar.gz file')

            # make copy of source tarball in deb_dist/
            local_source_tarball = os.path.split(source_tarball)[-1]
            shutil.copy2(source_tarball, local_source_tarball)
            source_tarball = local_source_tarball
            self.use_premade_distfile = source_tarball
        else:
            source_tarball = self.use_premade_distfile

        # Copy source tree assuming that package-0.1.tar.gz contains
        # single top-level path 'package-0.1'. The contents of this
        # directory are then used.

        if os.path.exists(fullpath_repackaged_dirname):
            shutil.rmtree(fullpath_repackaged_dirname)

        tmpdir = tempfile.mkdtemp()
        expand_sdist_file(os.path.abspath(source_tarball),
                          cwd=tmpdir)
        expanded_base_files = os.listdir(tmpdir)
        assert len(expanded_base_files) == 1
        actual_package_dirname = expanded_base_files[0]
        shutil.move(os.path.join(tmpdir, actual_package_dirname),
                    fullpath_repackaged_dirname)

        # ensure premade sdist can actually be used
        self.use_premade_distfile = os.path.abspath(self.use_premade_distfile)
        expand_dir = os.path.join(self.dist_dir, 'tmp_sdist_dsc')
        cleanup_dirs.append(expand_dir)
        if os.path.exists(expand_dir):
            shutil.rmtree(expand_dir)
        if not os.path.exists(self.dist_dir):
            os.mkdir(self.dist_dir)
        os.mkdir(expand_dir)

        expand_sdist_file(self.use_premade_distfile, cwd=expand_dir)

        is_tgz = False
        if self.use_premade_distfile.lower().endswith('.tar.gz'):
            is_tgz = True

        # now the sdist package is expanded in expand_dir
        expanded_root_files = os.listdir(expand_dir)
        assert len(expanded_root_files) == 1
        distname_in_premade_distfile = expanded_root_files[0]
        debianized_dirname = repackaged_dirname
        original_dirname = os.path.split(distname_in_premade_distfile)[-1]
        do_repack = False
        if is_tgz:
            source_tarball = self.use_premade_distfile
        else:
            log.warn('WARNING: .orig.tar.gz will be generated from sdist '
                     'archive ("%s") because it is not a .tar.gz file',
                     self.use_premade_distfile)
            do_repack = True

        if do_repack:
            tmp_dir = os.path.join(self.dist_dir, 'tmp_repacking_dir')
            os.makedirs(tmp_dir)
            cleanup_dirs.append(tmp_dir)
            source_tarball = os.path.join(tmp_dir, 'repacked_sdist.tar.gz')
            repack_tarball_with_debianized_dirname(self.use_premade_distfile,
                                                   source_tarball,
                                                   debianized_dirname,
                                                   original_dirname)
        if source_tarball is not None:
            # Because we deleted all .pyc files above, if the
            # original source dist has them, we will have
            # (wrongly) deleted them. So, quit loudly rather
            # than fail silently.
            for root, dirs, files in os.walk(fullpath_repackaged_dirname):
                for name in files:
                    if name.endswith('.pyc'):
                        raise RuntimeError('original source dist cannot '
                                           'contain .pyc files')

        ###############################################
        # 3. Find all directories

        for pkgdir in self.distribution.packages or []:
            debinfo.dirlist += ' ' + pkgdir.replace('.', '/')

        ###############################################
        # 4. Build source tree and rename it to be in self.dist_dir

        build_dsc(debinfo,
                  self.dist_dir,
                  repackaged_dirname,
                  orig_sdist=source_tarball,
                  patch_posix=self.patch_posix,
                  remove_expanded_source_dir=self.remove_expanded_source_dir,
                  sign_dsc=self.sign_results,
                  sign_key=self.sign_key,
                  ignore_source_changes=self.ignore_source_changes,
                  )

        for rmdir in cleanup_dirs:
            shutil.rmtree(rmdir)



import os
import stdeb.util as util

from distutils.core import Command

__all__ = ['bdist_deb']


class bdist_deb(Command):
    description = 'distutils command to create debian binary package'

    user_options = [
        ('sign-results', None,
         'Use gpg to sign the resulting .dsc and .changes file'),
        ('ignore-source-changes', None,
         'Ignore all changes on source when building source package '
         '(add -i.* option to dpkg-source'),
        ]
    boolean_options = [
        'sign-results',
        'ignore-source-changes',
        ]

    def initialize_options(self):
        self.sign_results = False
        self.ignore_source_changes = False

    def finalize_options(self):
        self.sign_results = bool(self.sign_results)
        self.ignore_source_changes = bool(self.ignore_source_changes)

    def run(self):
        # generate .dsc source pkg
        self.run_command('sdist_dsc')

        # get relevant options passed to sdist_dsc
        sdist_dsc = self.get_finalized_command('sdist_dsc')
        dsc_tree = sdist_dsc.dist_dir

        # execute system command and read output
        # (execute and read output of find cmd)
        target_dirs = []
        for entry in os.listdir(dsc_tree):
            fulldir = os.path.join(dsc_tree, entry)
            if os.path.isdir(fulldir):
                if entry == 'tmp_py2dsc':
                    continue
                target_dirs.append(fulldir)

        if len(target_dirs) > 1:
            raise ValueError('More than one directory in deb_dist. '
                             'Unsure which is source directory. All: %r' % (
                                target_dirs,))

        if len(target_dirs) == 0:
            raise ValueError('could not find debian source directory')

        # define system command to execute (gen .deb binary pkg)
        syscmd = ['dpkg-buildpackage', '-rfakeroot', '-b']

        if not self.sign_results:
            syscmd.append('-uc')

        if self.ignore_source_changes:
            syscmd.append('-i.*')

        util.process_command(syscmd, cwd=target_dirs[0])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from distutils.core import setup
from distutils.command.build import build
import sys


class my_build(build):
    """ensure (at runtime) we are running python 2"""
    def __init__(self, *args, **kwargs):
        assert sys.version_info[0] == 2
        build.__init__(self, *args, **kwargs)


setup(
    name='py2_only_pkg',
    packages=['py2_only_pkg'],
    version='0.1',
    cmdclass={'build': my_build},
    author='Mister Unicodé',
    author_email='mister.unicode@example.tld',
    description='Python 2 package with Unicodé fields',
    long_description='This is a Python 2 package with Unicodé data.',
)



import logging
__version__ = '0.11.0'  # keep in sync with ../setup.py

log = logging.getLogger('stdeb')
log.setLevel(logging.INFO)
handler = logging.StreamHandler()
handler.setLevel(logging.INFO)
formatter = logging.Formatter('%(message)s')
handler.setFormatter(formatter)
log.addHandler(handler)



from distutils.core import Command  # noqa: F401
from stdeb.command.common import common_debian_package_command

from stdeb.util import build_dsc, stdeb_cmdline_opts, \
     stdeb_cmd_bool_opts, stdeb_cfg_options


class debianize(common_debian_package_command):
    description = "distutils command to create a debian directory"

    user_options = stdeb_cmdline_opts + stdeb_cfg_options
    boolean_options = stdeb_cmd_bool_opts

    def run(self):
        debinfo = self.get_debinfo()
        if debinfo.patch_file != '':
            raise RuntimeError(
                'Patches cannot be applied in debianize command')

        dist_dir = None
        repackaged_dirname = None

        build_dsc(debinfo,
                  dist_dir,
                  repackaged_dirname,
                  debian_dir_only=True,
                  )



from __future__ import print_function
import os
import shutil
import subprocess
import sys
from configparser import ConfigParser  # noqa: F401
from distutils.util import strtobool
from distutils.fancy_getopt import FancyGetopt, translate_longopt
from stdeb.util import stdeb_cmdline_opts, stdeb_cmd_bool_opts
from stdeb.util import expand_sdist_file, apply_patch
from stdeb import log

from pkg_resources import Requirement, Distribution  # noqa: F401


class OptObj:
    pass


def runit(cmd, usage):
    if cmd not in ['sdist_dsc', 'bdist_deb']:
        raise ValueError('unknown command %r' % cmd)
    # process command-line options
    bool_opts = list(map(translate_longopt, stdeb_cmd_bool_opts))
    parser = FancyGetopt(stdeb_cmdline_opts+[
        ('help', 'h', "show detailed help message"),
        ])
    optobj = OptObj()
    args = parser.getopt(object=optobj)
    for option in optobj.__dict__:
        value = getattr(optobj, option)
        is_string = type(value) == str
        if option in bool_opts and is_string:
            setattr(optobj, option, strtobool(value))

    if hasattr(optobj, 'help'):
        print(usage)
        parser.set_option_table(stdeb_cmdline_opts)
        parser.print_help("Options:")
        return 0

    if len(args) != 1:
        log.error('not given single argument (distfile), args=%r', args)
        print(usage)
        return 1

    sdist_file = args[0]

    final_dist_dir = optobj.__dict__.get('dist_dir', 'deb_dist')
    tmp_dist_dir = os.path.join(final_dist_dir, 'tmp_py2dsc')
    if os.path.exists(tmp_dist_dir):
        shutil.rmtree(tmp_dist_dir)
    os.makedirs(tmp_dist_dir)

    if not os.path.isfile(sdist_file):
        log.error("Package %s not found." % sdist_file)
        sys.exit(1)

    patch_file = optobj.__dict__.get('patch_file', None)
    patch_level = int(optobj.__dict__.get('patch_level', 0))
    patch_posix = int(optobj.__dict__.get('patch_posix', 0))

    expand_dir = os.path.join(tmp_dist_dir, 'stdeb_tmp')
    if os.path.exists(expand_dir):
        shutil.rmtree(expand_dir)
    if not os.path.exists(tmp_dist_dir):
        os.mkdir(tmp_dist_dir)
    os.mkdir(expand_dir)

    expand_sdist_file(os.path.abspath(sdist_file), cwd=expand_dir)

    # now the sdist package is expanded in expand_dir
    expanded_root_files = os.listdir(expand_dir)
    assert len(expanded_root_files) == 1
    repackaged_dirname = expanded_root_files[0]
    fullpath_repackaged_dirname = os.path.join(
        tmp_dist_dir, repackaged_dirname)
    base_dir = os.path.join(expand_dir, expanded_root_files[0])
    if os.path.exists(fullpath_repackaged_dirname):
        # prevent weird build errors if this dir exists
        shutil.rmtree(fullpath_repackaged_dirname)
    os.renames(base_dir, fullpath_repackaged_dirname)
    del base_dir  # no longer useful

    ##############################################
    if patch_file is not None:
        log.info('py2dsc applying patch %s', patch_file)
        apply_patch(patch_file,
                    posix=patch_posix,
                    level=patch_level,
                    cwd=fullpath_repackaged_dirname)
        patch_already_applied = 1
    else:
        patch_already_applied = 0
    ##############################################

    abs_dist_dir = os.path.abspath(final_dist_dir)

    extra_args = []
    for long in parser.long_opts:
        if long in ['dist-dir=', 'patch-file=']:
            continue  # dealt with by this invocation
        attr = parser.get_attr_name(long).rstrip('=')
        if hasattr(optobj, attr):
            val = getattr(optobj, attr)
            if attr == 'extra_cfg_file':
                val = os.path.abspath(val)
            if long in bool_opts or long.replace('-', '_') in bool_opts:
                extra_args.append('--%s' % long)
            else:
                extra_args.append('--'+long+str(val))

    if patch_already_applied == 1:
        extra_args.append('--patch-already-applied')

    if cmd == 'bdist_deb':
        extra_args.append('bdist_deb')

    args = [sys.executable, 'setup.py', '--command-packages', 'stdeb.command',
            'sdist_dsc', '--dist-dir=%s' % abs_dist_dir,
            '--use-premade-distfile=%s' % os.path.abspath(sdist_file)
            ] + extra_args

    log.info('-='*35 + '-')
#    print >> sys.stderr, '-='*20
#    print >> sys.stderr, "Note that the .cfg file(s), if present, have not "\
#          "been read at this stage. If options are necessary, pass them "\
#          "from the command line"
    log.info("running the following command in directory: %s\n%s",
             fullpath_repackaged_dirname, ' '.join(args))
    log.info('-='*35 + '-')

    try:
        returncode = subprocess.call(
            args, cwd=fullpath_repackaged_dirname,
            )
    except Exception:
        log.error('ERROR running: %s', ' '.join(args))
        log.error('ERROR in %s', fullpath_repackaged_dirname)
        raise

    if returncode:
        log.error('ERROR running: %s', ' '.join(args))
        log.error('ERROR in %s', fullpath_repackaged_dirname)
        # log.error('   stderr: %s'res.stderr.read())
        # print >> sys.stderr, 'ERROR running: %s'%(' '.join(args),)
        # print >> sys.stderr, res.stderr.read()
        return returncode
        # raise RuntimeError('returncode %d'%returncode)
    # result = res.stdout.read().strip()

    shutil.rmtree(tmp_dist_dir)
    return returncode



import glob
import os
import stdeb.util as util

from distutils.core import Command

__all__ = ['install_deb']


class install_deb(Command):
    description = 'distutils command to install debian binary package'

    user_options = []
    boolean_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        # generate .deb file
        self.run_command('bdist_deb')

        # get relevant options passed to sdist_dsc
        sdist_dsc = self.get_finalized_command('sdist_dsc')

        # execute system command and read output
        # (execute and read output of find cmd)
        target_debs = glob.glob(os.path.join(sdist_dsc.dist_dir, '*.deb'))

        if len(target_debs) == 0:
            raise ValueError('could not find .deb file')

        for target_deb in target_debs:
            # define system command to execute (install .deb binary pkg)
            syscmd = ['dpkg', '--install', target_deb]
            util.process_command(syscmd)



import os
import sys

from stdeb import log
from distutils.core import Command
from distutils.errors import DistutilsModuleError

from stdeb.util import DebianInfo, DH_DEFAULT_VERS, stdeb_cfg_options


class common_debian_package_command(Command):
    def initialize_options(self):
        self.patch_already_applied = 0
        self.remove_expanded_source_dir = 0
        self.patch_posix = 0
        self.dist_dir = None
        self.extra_cfg_file = None
        self.patch_file = None
        self.patch_level = None
        self.ignore_install_requires = None
        self.debian_version = None
        self.no_backwards_compatibility = None
        self.guess_conflicts_provides_replaces = None
        if sys.version_info[0] == 2:
            self.with_python2 = 'True'
            self.with_python3 = 'False'
        else:
            assert sys.version_info[0] == 3
            self.with_python2 = 'False'
            self.with_python3 = 'True'
        self.no_python2_scripts = 'False'
        self.no_python3_scripts = 'False'
        self.force_x_python3_version = False
        self.allow_virtualenv_install_location = False
        self.with_dh_virtualenv = False
        self.with_dh_systemd = False
        self.sign_results = False
        self.sign_key = None
        self.ignore_source_changes = False
        self.compat = DH_DEFAULT_VERS

        # deprecated options
        self.default_distribution = None
        self.default_maintainer = None

        # make distutils happy by filling in default values
        for longopt, shortopt, description in stdeb_cfg_options:
            assert longopt.endswith('=')
            name = longopt[:-1]
            name = name.replace('-', '_')
            setattr(self, name, None)

    def finalize_options(self):
        def str_to_bool(mystr):
            if mystr.lower() == 'false':
                return False
            elif mystr.lower() == 'true':
                return True
            else:
                raise ValueError(
                    'bool string "%s" is not "true" or "false"' % mystr)
        if self.dist_dir is None:
            self.dist_dir = 'deb_dist'
        if self.patch_level is not None:
            self.patch_level = int(self.patch_level)

        if self.guess_conflicts_provides_replaces is None:
            # the default
            self.guess_conflicts_provides_replaces = False
        else:
            self.guess_conflicts_provides_replaces = str_to_bool(
                self.guess_conflicts_provides_replaces)

        self.with_python2 = str_to_bool(self.with_python2)
        self.with_python3 = str_to_bool(self.with_python3)
        self.no_python2_scripts = str_to_bool(self.no_python2_scripts)
        self.no_python3_scripts = str_to_bool(self.no_python3_scripts)
        if self.maintainer is not None:
            # Get the locale specifying the encoding in sys.argv
            import codecs
            import locale
            fs_enc = codecs.lookup(locale.getpreferredencoding()).name
            if hasattr(os, 'fsencode'):  # this exists only in Python 3
                m = os.fsencode(self.maintainer)  # convert to orig raw bytes

                # Now, convert these raw bytes into unicode.
                m = m.decode(fs_enc)  # Set your locale if you get errors here

                self.maintainer = m
            else:
                # Python 2
                if hasattr(self.maintainer, 'decode'):
                    self.maintainer = self.maintainer.decode(fs_enc)

    def get_debinfo(self):
        ###############################################
        # 1. setup initial variables
        #    A. create config defaults
        module_name = self.distribution.get_name()

        if 1:
            # set default maintainer
            if os.environ.get('DEBEMAIL'):
                guess_maintainer = "%s <%s>" % (
                    os.environ.get('DEBFULLNAME', os.environ['DEBEMAIL']),
                    os.environ['DEBEMAIL'])
            elif (
                self.distribution.get_maintainer() != 'UNKNOWN' and
                self.distribution.get_maintainer_email() != 'UNKNOWN'
            ):
                guess_maintainer = "%s <%s>" % (
                    self.distribution.get_maintainer(),
                    self.distribution.get_maintainer_email())
            elif (self.distribution.get_author() != 'UNKNOWN' and
                  self.distribution.get_author_email() != 'UNKNOWN'):
                guess_maintainer = "%s <%s>" % (
                    self.distribution.get_author(),
                    self.distribution.get_author_email())
            else:
                guess_maintainer = "unknown <unknown@unknown>"
        if self.default_maintainer is not None:
            log.warn('Deprecation warning: you are using the '
                     '--default-maintainer option. '
                     'Switch to the --maintainer option.')
            guess_maintainer = self.default_maintainer
        if hasattr(guess_maintainer, 'decode'):
            # python 2 : convert (back to) unicode
            guess_maintainer = guess_maintainer.decode('utf-8')

        #    B. find config files (if any)
        cfg_files = []
        if self.extra_cfg_file is not None:
            cfg_files.append(self.extra_cfg_file)

        use_setuptools = True
        try:
            ei_cmd = self.distribution.get_command_obj('egg_info')
        except DistutilsModuleError:
            use_setuptools = False

        config_fname = 'stdeb.cfg'
        # Distutils fails if not run from setup.py dir, so this is OK.
        if os.path.exists(config_fname):
            cfg_files.append(config_fname)

        if use_setuptools:
            self.run_command('egg_info')
            egg_info_dirname = ei_cmd.egg_info

            # Pickup old location of stdeb.cfg
            config_fname = os.path.join(egg_info_dirname, 'stdeb.cfg')
            if os.path.exists(config_fname):
                log.warn('Deprecation warning: stdeb detected old location of '
                         'stdeb.cfg in %s. This file will be used, but you '
                         'should move it alongside setup.py.' %
                         egg_info_dirname)
                cfg_files.append(config_fname)

            egg_module_name = egg_info_dirname[
                :egg_info_dirname.index('.egg-info')]
            egg_module_name = egg_module_name.split(os.sep)[-1]
        else:
            # We don't have setuptools, so guess egg_info_dirname to
            # find old stdeb.cfg.

            entries = os.listdir(os.curdir)
            for entry in entries:
                if not (entry.endswith('.egg-info') and os.path.isdir(entry)):
                    continue
                # Pickup old location of stdeb.cfg
                config_fname = os.path.join(entry, 'stdeb.cfg')
                if os.path.exists(config_fname):
                    log.warn('Deprecation warning: stdeb detected '
                             'stdeb.cfg in %s. This file will be used, but '
                             'you should move it alongside setup.py.' % entry)
                    cfg_files.append(config_fname)

        upstream_version = self.distribution.get_version()
        bad_chars = ':_'
        for bad_char in bad_chars:
            if bad_char in upstream_version:
                raise ValueError("Illegal character (%r) detected in version. "
                                 "This will break the debian tools." %
                                 bad_char)

        description = self.distribution.get_description()
        if hasattr(description, 'decode'):
            # python 2 : convert (back to) unicode
            description = description.decode('utf-8')
        description = description[:60]

        long_description = self.distribution.get_long_description()
        if hasattr(long_description, 'decode'):
            # python 2 : convert (back to) unicode
            long_description = long_description.decode('utf-8')
        long_description = long_description

        debinfo = DebianInfo(
            cfg_files=cfg_files,
            module_name=module_name,
            default_distribution=self.default_distribution,
            guess_maintainer=guess_maintainer,
            upstream_version=upstream_version,
            has_ext_modules=self.distribution.has_ext_modules(),
            description=description,
            long_description=long_description,
            homepage=self.distribution.get_url(),
            patch_file=self.patch_file,
            patch_level=self.patch_level,
            debian_version=self.debian_version,
            setup_requires=(),  # XXX How do we get the setup_requires?
            use_setuptools=use_setuptools,
            guess_conflicts_provides_replaces=(
                self.guess_conflicts_provides_replaces),
            sdist_dsc_command=self,
            with_python2=self.with_python2,
            with_python3=self.with_python3,
            no_python2_scripts=self.no_python2_scripts,
            no_python3_scripts=self.no_python3_scripts,
            force_x_python3_version=self.force_x_python3_version,
            allow_virtualenv_install_location=(
                self.allow_virtualenv_install_location),
            compat=self.compat,
            with_dh_virtualenv=self.with_dh_virtualenv,
            with_dh_systemd=self.with_dh_systemd,
        )
        return debinfo



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from distutils.core import setup

setup(name='simple_pkg',
      packages=['simple_pkg'],
      version='0.1',
      author='Mister Unicodé',
      author_email='mister.unicode@example.tld',
      description='Python package with Unicodé fields',
      long_description='This is a Python package with Unicodé data.',
      )



# -*- coding: utf-8 -*-
"""
A replacement transport for Python xmlrpc library.

Usage:

    >>> import xmlrpclib
    >>> from transport import RequestsTransport
    >>> s = xmlrpclib.ServerProxy(
    >>>     'http://yoursite.com/xmlrpc', transport=RequestsTransport())
    >>> s.demo.sayHello()
    Hello!
"""
try:
    import xmlrpc.client as xmlrpc
except ImportError:
    import xmlrpclib as xmlrpc

import requests
import requests.utils

import sys
from distutils.version import StrictVersion
import warnings


class RequestsTransport(xmlrpc.Transport):
    """
    Drop in Transport for xmlrpclib that uses Requests instead of httplib
    """
    # change our user agent to reflect Requests
    user_agent = "Python XMLRPC with Requests (python-requests.org)"

    # override this if you'd like to https
    use_https = False

    def request(self, host, handler, request_body, verbose):
        """
        Make an xmlrpc request.
        """
        headers = {'User-Agent': self.user_agent,
                   'Content-Type': 'text/xml',
                   }
        url = self._build_url(host, handler)
        kwargs = {}
        if StrictVersion(requests.__version__) >= StrictVersion('0.8.8'):
            kwargs['verify'] = True
        else:
            if self.use_https:
                warnings.warn(
                    'using https transport but no certificate '
                    'verification. (Hint: upgrade requests package.)')
        try:
            resp = requests.post(url, data=request_body, headers=headers,
                                 **kwargs)
        except ValueError:
            raise
        except Exception:
            raise  # something went wrong
        else:
            try:
                resp.raise_for_status()
            except requests.RequestException as e:
                raise xmlrpc.ProtocolError(
                    url, resp.status_code, str(e), resp.headers)
            else:
                return self.parse_response(resp)

    def parse_response(self, resp):
        """
        Parse the xmlrpc response.
        """
        p, u = self.getparser()

        if hasattr(resp, 'text'):
            # modern requests will do this for us
            text = resp.text  # this is unicode(py2)/str(py3)
        else:

            encoding = requests.utils.get_encoding_from_headers(resp.headers)
            if encoding is None:
                encoding = 'utf-8'  # FIXME: what to do here?

            if sys.version_info[0] == 2:
                text = unicode(  # noqa: F821
                    resp.content, encoding, errors='replace')
            else:
                assert sys.version_info[0] == 3
                text = str(resp.content, encoding, errors='replace')
        p.feed(text)
        p.close()
        return u.close()

    def _build_url(self, host, handler):
        """
        Build a url for our request based on the host, handler and use_http
        property
        """
        scheme = 'https' if self.use_https else 'http'
        return '%s://%s/%s' % (scheme, host, handler)



from distutils.core import setup
import codecs

with codecs.open('README.rst', encoding='utf-8') as file:
    long_description = file.read()


setup(
    name='stdeb',
    # Keep version in sync with stdeb/__init__.py and install section
    # of README.rst.
    version='0.11.0',
    author='Andrew Straw',
    author_email='strawman@astraw.com',
    description='Python to Debian source package conversion utility',
    long_description=long_description,
    license='MIT',
    url='http://github.com/astraw/stdeb',
    packages=['stdeb', 'stdeb.command'],
    scripts=[
        'scripts/py2dsc',
        'scripts/py2dsc-deb',
        'scripts/pypi-download',
        'scripts/pypi-install',
    ],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: MIT License',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
    ],
)
